/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __NAUTILUS_FREEDESKTOP_GENERATED_H__
#define __NAUTILUS_FREEDESKTOP_GENERATED_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.FileManager1 */

#define NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1 (nautilus_freedesktop_file_manager1_get_type ())
#define NAUTILUS_FREEDESKTOP_FILE_MANAGER1(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1, NautilusFreedesktopFileManager1))
#define NAUTILUS_FREEDESKTOP_IS_FILE_MANAGER1(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1))
#define NAUTILUS_FREEDESKTOP_FILE_MANAGER1_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1, NautilusFreedesktopFileManager1Iface))

struct _NautilusFreedesktopFileManager1;
typedef struct _NautilusFreedesktopFileManager1 NautilusFreedesktopFileManager1;
typedef struct _NautilusFreedesktopFileManager1Iface NautilusFreedesktopFileManager1Iface;

struct _NautilusFreedesktopFileManager1Iface
{
  GTypeInterface parent_iface;


  gboolean (*handle_show_folders) (
    NautilusFreedesktopFileManager1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_URIs,
    const gchar *arg_StartupId);

  gboolean (*handle_show_item_properties) (
    NautilusFreedesktopFileManager1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_URIs,
    const gchar *arg_StartupId);

  gboolean (*handle_show_items) (
    NautilusFreedesktopFileManager1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_URIs,
    const gchar *arg_StartupId);

  const gchar *const * (*get_open_locations) (NautilusFreedesktopFileManager1 *object);

  GVariant * (*get_open_windows_with_locations) (NautilusFreedesktopFileManager1 *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (NautilusFreedesktopFileManager1, g_object_unref)
#endif

GType nautilus_freedesktop_file_manager1_get_type (void);

GDBusInterfaceInfo *nautilus_freedesktop_file_manager1_interface_info (void);
guint nautilus_freedesktop_file_manager1_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void nautilus_freedesktop_file_manager1_complete_show_folders (
    NautilusFreedesktopFileManager1 *object,
    GDBusMethodInvocation *invocation);

void nautilus_freedesktop_file_manager1_complete_show_items (
    NautilusFreedesktopFileManager1 *object,
    GDBusMethodInvocation *invocation);

void nautilus_freedesktop_file_manager1_complete_show_item_properties (
    NautilusFreedesktopFileManager1 *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void nautilus_freedesktop_file_manager1_call_show_folders (
    NautilusFreedesktopFileManager1 *proxy,
    const gchar *const *arg_URIs,
    const gchar *arg_StartupId,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_freedesktop_file_manager1_call_show_folders_finish (
    NautilusFreedesktopFileManager1 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_freedesktop_file_manager1_call_show_folders_sync (
    NautilusFreedesktopFileManager1 *proxy,
    const gchar *const *arg_URIs,
    const gchar *arg_StartupId,
    GCancellable *cancellable,
    GError **error);

void nautilus_freedesktop_file_manager1_call_show_items (
    NautilusFreedesktopFileManager1 *proxy,
    const gchar *const *arg_URIs,
    const gchar *arg_StartupId,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_freedesktop_file_manager1_call_show_items_finish (
    NautilusFreedesktopFileManager1 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_freedesktop_file_manager1_call_show_items_sync (
    NautilusFreedesktopFileManager1 *proxy,
    const gchar *const *arg_URIs,
    const gchar *arg_StartupId,
    GCancellable *cancellable,
    GError **error);

void nautilus_freedesktop_file_manager1_call_show_item_properties (
    NautilusFreedesktopFileManager1 *proxy,
    const gchar *const *arg_URIs,
    const gchar *arg_StartupId,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_freedesktop_file_manager1_call_show_item_properties_finish (
    NautilusFreedesktopFileManager1 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_freedesktop_file_manager1_call_show_item_properties_sync (
    NautilusFreedesktopFileManager1 *proxy,
    const gchar *const *arg_URIs,
    const gchar *arg_StartupId,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
const gchar *const *nautilus_freedesktop_file_manager1_get_open_locations (NautilusFreedesktopFileManager1 *object);
gchar **nautilus_freedesktop_file_manager1_dup_open_locations (NautilusFreedesktopFileManager1 *object);
void nautilus_freedesktop_file_manager1_set_open_locations (NautilusFreedesktopFileManager1 *object, const gchar *const *value);

GVariant *nautilus_freedesktop_file_manager1_get_open_windows_with_locations (NautilusFreedesktopFileManager1 *object);
GVariant *nautilus_freedesktop_file_manager1_dup_open_windows_with_locations (NautilusFreedesktopFileManager1 *object);
void nautilus_freedesktop_file_manager1_set_open_windows_with_locations (NautilusFreedesktopFileManager1 *object, GVariant *value);


/* ---- */

#define NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_PROXY (nautilus_freedesktop_file_manager1_proxy_get_type ())
#define NAUTILUS_FREEDESKTOP_FILE_MANAGER1_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_PROXY, NautilusFreedesktopFileManager1Proxy))
#define NAUTILUS_FREEDESKTOP_FILE_MANAGER1_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_PROXY, NautilusFreedesktopFileManager1ProxyClass))
#define NAUTILUS_FREEDESKTOP_FILE_MANAGER1_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_PROXY, NautilusFreedesktopFileManager1ProxyClass))
#define NAUTILUS_FREEDESKTOP_IS_FILE_MANAGER1_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_PROXY))
#define NAUTILUS_FREEDESKTOP_IS_FILE_MANAGER1_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_PROXY))

typedef struct _NautilusFreedesktopFileManager1Proxy NautilusFreedesktopFileManager1Proxy;
typedef struct _NautilusFreedesktopFileManager1ProxyClass NautilusFreedesktopFileManager1ProxyClass;
typedef struct _NautilusFreedesktopFileManager1ProxyPrivate NautilusFreedesktopFileManager1ProxyPrivate;

struct _NautilusFreedesktopFileManager1Proxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  NautilusFreedesktopFileManager1ProxyPrivate *priv;
};

struct _NautilusFreedesktopFileManager1ProxyClass
{
  GDBusProxyClass parent_class;
};

GType nautilus_freedesktop_file_manager1_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (NautilusFreedesktopFileManager1Proxy, g_object_unref)
#endif

void nautilus_freedesktop_file_manager1_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
NautilusFreedesktopFileManager1 *nautilus_freedesktop_file_manager1_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
NautilusFreedesktopFileManager1 *nautilus_freedesktop_file_manager1_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void nautilus_freedesktop_file_manager1_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
NautilusFreedesktopFileManager1 *nautilus_freedesktop_file_manager1_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
NautilusFreedesktopFileManager1 *nautilus_freedesktop_file_manager1_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_SKELETON (nautilus_freedesktop_file_manager1_skeleton_get_type ())
#define NAUTILUS_FREEDESKTOP_FILE_MANAGER1_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_SKELETON, NautilusFreedesktopFileManager1Skeleton))
#define NAUTILUS_FREEDESKTOP_FILE_MANAGER1_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_SKELETON, NautilusFreedesktopFileManager1SkeletonClass))
#define NAUTILUS_FREEDESKTOP_FILE_MANAGER1_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_SKELETON, NautilusFreedesktopFileManager1SkeletonClass))
#define NAUTILUS_FREEDESKTOP_IS_FILE_MANAGER1_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_SKELETON))
#define NAUTILUS_FREEDESKTOP_IS_FILE_MANAGER1_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), NAUTILUS_FREEDESKTOP_TYPE_FILE_MANAGER1_SKELETON))

typedef struct _NautilusFreedesktopFileManager1Skeleton NautilusFreedesktopFileManager1Skeleton;
typedef struct _NautilusFreedesktopFileManager1SkeletonClass NautilusFreedesktopFileManager1SkeletonClass;
typedef struct _NautilusFreedesktopFileManager1SkeletonPrivate NautilusFreedesktopFileManager1SkeletonPrivate;

struct _NautilusFreedesktopFileManager1Skeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  NautilusFreedesktopFileManager1SkeletonPrivate *priv;
};

struct _NautilusFreedesktopFileManager1SkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType nautilus_freedesktop_file_manager1_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (NautilusFreedesktopFileManager1Skeleton, g_object_unref)
#endif

NautilusFreedesktopFileManager1 *nautilus_freedesktop_file_manager1_skeleton_new (void);


G_END_DECLS

#endif /* __NAUTILUS_FREEDESKTOP_GENERATED_H__ */
